// DLL SBASIC : soumission requete HTTP
// Author: Dominique Ezvan

#include <windows.h>
#include "sbasic.h"
#include <wininet.h>

#define MAX_URL     256
#define MAX_BUFFER  1024
#define BUFFER_SIZE 65536

static HINTERNET InternetHandle;
static PSBASIC_UTIL SaveSbasicUtil;
static ULONG BufferCount;
static ULONG Code;
static PUCHAR *BufferPtr;
static enum {Utf1, Utf2, Utf3, Utf4} Utf8Mode;
static enum {Byte2, Byte3, Byte4} Utf8State;
static ULONG UniCode[] = {0x20AC, 0x0081, 0x201A, 0x0192, 0x201E, 0x2026, 0x2020, 0x2021,
						  0x02C6, 0x2030, 0x0160, 0x2039, 0x0152, 0x008D, 0x017D, 0x008F,
						  0x0090, 0x2018, 0x2019, 0x201C, 0x201D, 0x2022, 0x2013, 0x2014,
						  0x02DC, 0x2122, 0x0161, 0x203A, 0x0153, 0x009D, 0x017E, 0x0178,
						  0x00A0, 0x00A1, 0x00A2, 0x00A3, 0x00A4, 0x00A5, 0x00A6, 0x00A7,
						  0x00A8, 0x00A9, 0x00AA, 0x00AB, 0x00AC, 0x00AD, 0x00AE, 0x00AF,
						  0x00B0, 0x00B1, 0x00B2, 0x00B3, 0x00B4, 0x00B5, 0x00B6, 0x00B7,
						  0x00B8, 0x00B9, 0x00BA, 0x00BB, 0x00BC, 0x00BD, 0x00BE, 0x00BF,
						  0x00C0, 0x00C1, 0x00C2, 0x00C3, 0x00C4, 0x00C5, 0x00C6, 0x00C7,
						  0x00C8, 0x00C9, 0x00CA, 0x00CB, 0x00CC, 0x00CD, 0x00CE, 0x00CF,
						  0x00D0, 0x00D1, 0x00D2, 0x00D3, 0x00D4, 0x00D5, 0x00D6, 0x00D7,
						  0x00D8, 0x00D9, 0x00DA, 0x00DB, 0x00DC, 0x00DD, 0x00DE, 0x00DF,
						  0x00E0, 0x00E1, 0x00E2, 0x00E3, 0x00E4, 0x00E5, 0x00E6, 0x00E7,
						  0x00E8, 0x00E9, 0x00EA, 0x00EB, 0x00EC, 0x00ED, 0x00EE, 0x00EF,
						  0x00F0, 0x00F1, 0x00F2, 0x00F3, 0x00F4, 0x00F5, 0x00F6, 0x00F7,
						  0x00F8, 0x00F9, 0x00FA, 0x00FB, 0x00FC, 0x00FD, 0x00FE, 0x00FF};
static UCHAR OemFixedFont[] = {0x3F, 0x3F, 0x3F, 0x9F, 0x3F, 0x3F, 0x3F, 0x3F, 0x3F, 0x3F, 0x3F, 0x3C, 0x3F, 0x3F, 0x3F, 0x3F,
							  0x3F, 0x60, 0x27, 0x22, 0x22, 0xF7, 0xF0, 0xC4, 0x7E, 0x3F, 0x3F, 0x3E, 0x3F, 0x3F, 0x3F, 0x98,
							  0x20, 0xAD, 0xBD, 0x9C, 0xCF, 0xBE, 0xDD, 0xF5, 0xF9, 0xB8, 0xA6, 0xAE, 0xAA, 0xF0, 0xA9, 0x20,
							  0xF8, 0xF1, 0xFD, 0xFC, 0x27, 0xE6, 0xF4, 0xFA, 0x2C, 0xFB, 0xA7, 0xAF, 0xAC, 0xAB, 0xF3, 0xA8,
							  0xB7, 0xB5, 0xB6, 0xC7, 0x8E, 0x8F, 0x92, 0x80, 0xD4, 0x90, 0xD2, 0xD3, 0xDE, 0xD6, 0xD7, 0xD8,
							  0xD1, 0xA5, 0xE3, 0xE0, 0xE2, 0xE5, 0x99, 0x9E, 0x9D, 0xEB, 0xE9, 0xEA, 0x9A, 0xED, 0xE8, 0xE1,
							  0x85, 0xA0, 0x83, 0xC6, 0x84, 0x86, 0x91, 0x87, 0x8A, 0x82, 0x88, 0x89, 0x8D, 0xA1, 0x8C, 0x8B,
							  0xD0, 0xA4, 0x95, 0xA2, 0x93, 0xE4, 0x94, 0xF6, 0x9B, 0x97, 0xA3, 0x96, 0x81, 0xEC, 0xE7, 0x98};

// Conversion UTF8 vers OEM
static BOOLEAN Utf8ToOem(UCHAR CharIn, PUCHAR CharOut) {
	ULONG i;

	switch (Utf8Mode) {
	case Utf1:
		if ((CharIn & 0x80) == 0x00) {
			*CharOut = CharIn;
			return TRUE;
		}
		if ((CharIn & 0xE0) == 0xC0) {
			Code = (CharIn & 0x1F) << 6;
			Utf8Mode = Utf2;
			return FALSE;
		}
		if ((CharIn & 0xF0) == 0xE0) {
			Code = (CharIn & 0x0F) << 12;
			Utf8Mode = Utf3;
			Utf8State = Byte2;
			return FALSE;
		}
		if ((CharIn & 0xFC) == 0xF0) {
			Code = (CharIn & 0x03) << 18;
			Utf8Mode = Utf4;
			Utf8State = Byte2;
			return FALSE;
		}
		return FALSE;
	case Utf2:
		if ((CharIn & 0xC0) == 0x80) {
			Code |= CharIn & 0x3F;
ReturnOem:
			if (Code < 0x80)
				*CharOut = (UCHAR)Code;
			else {
				for (i = 0; i < 128; i++)
					if (Code == UniCode[i])
						break;
				if (i < 128)
					*CharOut = OemFixedFont[i];
				else
					*CharOut = '?';
			}
			Utf8Mode = Utf1;
			return TRUE;
		}
		goto Utf8Error;
	case Utf3:
		switch (Utf8State) {
		case Byte2:
			if ((CharIn & 0xC0) == 0x80) {
				Code |= (CharIn & 0x3F) << 6;
				Utf8State = Byte3;
				return FALSE;
			}
			goto Utf8Error;
		case Byte3:
			if ((CharIn & 0xC0) == 0x80) {
				Code |= CharIn & 0x3F;
				goto ReturnOem;
			}
			goto Utf8Error;
		}
	case Utf4:
		switch (Utf8State) {
		case Byte2:
			if ((CharIn & 0xC0) == 0x80) {
				Code |= (CharIn & 0x3F) << 12;
				Utf8State = Byte3;
				return FALSE;
			}
			goto Utf8Error;
		case Byte3:
			if ((CharIn & 0xC0) == 0x80) {
				Code |= (CharIn & 0x3F) << 6;
				Utf8State = Byte4;
				return FALSE;
			}
			goto Utf8Error;
		case Byte4:
			if ((CharIn & 0xC0) == 0x80) {
				Code |= CharIn & 0x3F;
				goto ReturnOem;
			}
			goto Utf8Error;
		}
	}
Utf8Error:
	Utf8Mode = Utf1;
	return FALSE;
}

// Ouverture connexion internet
__declspec(dllexport) UCHAR __cdecl HttpOpen(ULONG ArgCount, PSBASIC_UTIL SbasicUtil) {

	// Verification arguments
	if (ArgCount != 0)
		return err_bad_arg;

	if (InternetHandle == NULL) {
		InternetHandle = InternetOpen("HTTP Command Line Utility", INTERNET_OPEN_TYPE_PRECONFIG, NULL, NULL, 0);
		if (InternetHandle == NULL)
			return err_no_file;
	}
	return 0;
}

// Fermeture connexion internet
__declspec(dllexport) UCHAR __cdecl HttpClose(ULONG ArgCount, PSBASIC_UTIL SbasicUtil) {

	// Verification arguments
	if (ArgCount != 0)
		return err_bad_arg;

	if (InternetHandle != NULL) {
		InternetCloseHandle(InternetHandle);
		InternetHandle = NULL;
	}
	return 0;
}

// Soumission requette HTTP
__declspec(dllexport) UCHAR __cdecl HttpSubmit(ULONG ArgCount, PSBASIC_UTIL SbasicUtil, ULONG Arg1Type, PSBASIC_STRING Url, ULONG Arg2Type, PSBASIC_STRING Data, ULONG Arg3Type, PSBASIC_INTEGER Raw) {
	ULONG length;
	ULONG index;
	ULONG byteCount[MAX_BUFFER];
	ULONG i;
	ULONG j;
	PUCHAR dataBase;
	PUCHAR ptr1;
	PUCHAR ptr2;
	HINTERNET urlHandle;
	UCHAR c;
	UCHAR url[MAX_URL];

	SaveSbasicUtil = SbasicUtil;
	if (InternetHandle == NULL)
		return err_not_opened;

	// Verification arguments
	if (ArgCount != 3 || Arg1Type != TSBASIC_STRING || Arg2Type != TSBASIC_STRING || Arg3Type != TSBASIC_INTEGER)
		return err_bad_arg;

	// soumet requette
	length = Url->Length;
	if (length > MAX_URL - 1)
		length = MAX_URL - 1;
	CopyMemory(url, Url->Address, length);
	url[length] = 0;
	urlHandle = InternetOpenUrl(InternetHandle, url, NULL, 0, INTERNET_FLAG_NO_CACHE_WRITE | INTERNET_FLAG_NO_UI, 0);
	if (urlHandle == NULL)
		return err_file_open;

	// recupere les donnees
	index = 0;
	length = 0;
	while (index < MAX_BUFFER) {
		if (index >= BufferCount) {
			BufferPtr[BufferCount] = VirtualAlloc(NULL, BUFFER_SIZE, MEM_COMMIT, PAGE_READWRITE);
			if (BufferPtr[BufferCount] == NULL) {
				InternetCloseHandle(urlHandle);
				return err_memory_ovf;
			}
			BufferCount++;
		}
		if (!InternetReadFile(urlHandle, BufferPtr[index], BUFFER_SIZE, &byteCount[index])) {
			InternetCloseHandle(urlHandle);
			return err_read;
		}
		if (byteCount[index] == 0)
			break;
		length += byteCount[index];
		index++;
	}

	// fermeture handle internet
	InternetCloseHandle(urlHandle);

	// conversion UTF8 vers OEM
	if (*Raw == 0) {
		length = 0;
		Utf8Mode = Utf1;
		i = 0;
		while (i < index) {
			ptr1 = BufferPtr[i];
			j = 0;
			while (j < byteCount[i]) {
				if (Utf8ToOem(*ptr1++, &c))
					length++;
				j++;
			}
			i++;
		}
	}

	// return data
	dataBase = SbasicUtil->SbMemAlloc(length);
	ptr2 = dataBase;
	i = 0;
	if (*Raw != 0) {
		while (i < index) {
			CopyMemory(ptr2, BufferPtr[i], byteCount[i]);
			ptr2 += byteCount[i];
			i++;
		}
	} else {
		Utf8Mode = Utf1;
		i = 0;
		while (i < index) {
			ptr1 = BufferPtr[i];
			j = 0;
			while (j < byteCount[i]) {
				if (Utf8ToOem(*ptr1++, &c))
					*ptr2++ = c;
				j++;
			}
			i++;
		}
	}
	if (Data->Address != NULL && Data->Length != 0)
		SbasicUtil->SbMemFree(Data->Address, Data->Length);
	Data->Address = dataBase;
	Data->Length = length;

	return 0;
}

// Mise chaine au standard URL
VOID encode(UCHAR chr, PUCHAR *ptr, ULONG *length) {
	ULONG i;
	UCHAR c;

	if (*length < MAX_URL) {
		*(*ptr)++ = '%';
		(*length)++;
	}
	i = (chr & 0xF0) >> 4;
	if (i < 10)
		c = (UCHAR)('0' + i);
	else
		c = (UCHAR)('A' + i - 10);
	if (*length < MAX_URL) {
		*(*ptr)++ = c;
		(*length)++;
	}
	i = chr & 0x0F;
	if (i < 10)
		c = (UCHAR)('0' + i);
	else
		c = (UCHAR)('A' + i - 10);
	if (*length < MAX_URL) {
		*(*ptr)++ = c;
		(*length)++;
	}
}

__declspec(dllexport) UCHAR __cdecl HttpUrl(ULONG ArgCount, PSBASIC_UTIL SbasicUtil, ULONG Arg1Type, PSBASIC_STRING String) {
	ULONG length;
	ULONG i;
	ULONG j;
	ULONG k;
	ULONG u;
	PUCHAR address;
	PUCHAR ptr1;
	PUCHAR ptr2;
	UCHAR c;
	UCHAR url[MAX_URL];
	BOOLEAN update;

	SaveSbasicUtil = SbasicUtil;

	// Verification arguments
	if (ArgCount != 1 || Arg1Type != TSBASIC_STRING)
		return err_bad_arg;

	// Traitement chaine
	length = String->Length;
	ptr1 = String->Address;
	ptr2 = url;
	i = 0;
	j = 0;
	update = FALSE;
	while (i < length && j < MAX_URL) {
		c = *ptr1++;
		i++;
		if (c < 0x20) {
			update = TRUE;
			continue;
		}
		if (c == 0x20) {
			*ptr2++ = '+';
			j++;
			update = TRUE;
			continue;
		}
		if (c < 0x80) {
			if ((c >= 'A' && c <= 'Z') || (c >= 'a' && c <= 'z') || (c >= '0' && c <= '9') || c == '-' || c == '_' || c == '.' || c == '~') {
				*ptr2++ = c;
				j++;
				continue;
			}
			encode(c, &ptr2, &j);
			update = TRUE;
			continue;
		}
		k = 0;
		while (k < 128 && c != OemFixedFont[k])
			k++;
		if (k < 128) {
			u = UniCode[k];
			if (u < 0x0080)
				encode((UCHAR)u, &ptr2, &j);
			else if (u < 0x0800) {
				encode((UCHAR)(((u & 0x07C0) >> 6) | 0xC0), &ptr2, &j);
				encode((UCHAR)((u & 0x003F) | 0x80), &ptr2, &j);
			} else {
				encode((UCHAR)(((u & 0xF000) >> 12) | 0xE0), &ptr2, &j);
				encode((UCHAR)(((u & 0x0FC0) >> 6) | 0x80), &ptr2, &j);
				encode((UCHAR)((u & 0x03F) | 0x80), &ptr2, &j);
			}
		}
		update = TRUE;
	}
	if (update) {
		address = SbasicUtil->SbMemAlloc(j);
		CopyMemory(address, url, j);
		if (String->Address != NULL && String->Length != 0)
			SbasicUtil->SbMemFree(String->Address, String->Length);
		String->Address = address;
		String->Length = j;
	}

	return 0;
}

// Interface DLL
BOOL WINAPI DllMain(HINSTANCE hinstDLL, DWORD fdwReason, LPVOID lpReserved ) {

	// Perform actions based on the reason for calling.
    switch(fdwReason) { 
        case DLL_PROCESS_ATTACH:
         // Initialize once for each new process.
         // Return FALSE to fail DLL load.
			InternetHandle = NULL;
			BufferCount = 0;
			BufferPtr = VirtualAlloc(NULL, MAX_BUFFER * sizeof(PUCHAR), MEM_COMMIT, PAGE_READWRITE);
			if (BufferPtr == NULL)
				return FALSE; // Not successful
            break;

        case DLL_THREAD_ATTACH:
         // Do thread-specific initialization.
            break;

        case DLL_THREAD_DETACH:
         // Do thread-specific cleanup.
            break;

        case DLL_PROCESS_DETACH:
         // Perform any necessary cleanup.
			if (InternetHandle != NULL)
				InternetCloseHandle(InternetHandle);
			while (BufferCount != 0) {
				BufferCount--;
				VirtualFree(BufferPtr[BufferCount], BUFFER_SIZE, 0);
			}
            break;
    }
    return TRUE;  // Successful DLL_XXX.
}